#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <leitm.h>
#include "perlin.h"

#define PI		3.1415926535897931
#define TWO_PI	6.2831853071795862

static float cloud_curve(float v, float cloud_cover, float sharpness)
{
	float c, dens;
	
	c = cloud_cover - v;
	if(c < 0.0) c = 0.0;

	dens = 1.0 - pow(1.0 - sharpness, c);
	return dens;
}

static void cross(float *rx, float *ry, float *rz, float x1, float y1, float z1, float x2, float y2, float z2)
{
	*rx = y1 * z2 - z1 * y2;
	*ry = z1 * x2 - x1 * z2;
	*rz = x1 * y2 - y1 * x2;
}

static void normalize(float *x, float *y, float *z)
{
	float len = sqrt(*x * *x + *y * *y + *z * *z);
	*x /= len;
	*y /= len;
	*z /= len;
}

#define DUMP_FILE	"ndump"

#define NCACHE_CENTER(x, y, z)		pncache[z * xsz * ysz * 5 + y * xsz * 5 + x * 5]
#define NCACHE_LEFT(x, y, z)		pncache[z * xsz * ysz * 5 + y * xsz * 5 + x * 5 + 0]
#define NCACHE_RIGHT(x, y, z)		pncache[z * xsz * ysz * 5 + y * xsz * 5 + x * 5 + 1]
#define NCACHE_UP(x, y, z)			pncache[z * xsz * ysz * 5 + y * xsz * 5 + x * 5 + 2]
#define NCACHE_DOWN(x, y, z)		pncache[z * xsz * ysz * 5 + y * xsz * 5 + x * 5 + 3]

static float *pncache;
static int calc_noise = 1;

void invalidate_cache(void)
{
	free(pncache);
	pncache = 0;
	calc_noise = 1;
	remove(DUMP_FILE);
}

struct dump_hdr {
	int x, y, z;
	float sdist;
} __attribute__((packed));

#define MULT	16.0
void cloud_image(unsigned int *img, unsigned int *norm, int xsz, int ysz, int zsz)
{
	int x, y, z;
	size_t elem_count = xsz * ysz * zsz * 5;
	extern float cloud_cover, cloud_sharpness, slice_dist_fact;
	extern lmwidget *pbar_calc;

	if(!pncache) {
		FILE *fp;
		pncache = malloc(elem_count * sizeof *pncache);

		if((fp = fopen(DUMP_FILE, "rb"))) {
			struct dump_hdr hdr;

			if(fread(&hdr, sizeof hdr, 1, fp) &&
					hdr.x == xsz && hdr.y == ysz && hdr.z == zsz &&
					fabs(hdr.sdist - slice_dist_fact) < 0.001) {
				if(fread(pncache, sizeof *pncache, elem_count, fp) == elem_count) {
					fprintf(stderr, "noise dump used\n");
					calc_noise = 0;
				}
			}
			fclose(fp);
		}
	}

	fputc('\n', stderr);
	for(z=0; z<zsz; z++) {
		fprintf(stderr, "\rcalculating cloud frame: %d / %d", z, zsz);
		/*lm_progress_set_value(pbar_calc, (float)z / (float)zsz);*/

		for(y=0; y<ysz; y++) {
			for(x=0; x<xsz; x++) {
				float fx, fy, fz, pnoise, cloud;
				int val;

				fx = MULT * (float)x / (float)xsz;
				fy = MULT * (float)y / (float)ysz;
				fz = 0.1 + slice_dist_fact * (float)z / (float)zsz;

				if(calc_noise) {
					pnoise = perlin_noise_3d(fx, fy, fz, 2, 2, 4) * 0.5 + 0.5;
					NCACHE_CENTER(x, y, z) = pnoise;
				} else {
					pnoise = NCACHE_CENTER(x, y, z);
				}

				cloud = cloud_curve(pnoise, cloud_cover, cloud_sharpness);

				if(norm) {
					float h = (MULT / 2.0) / (float)xsz;
					float cleft, cright, cup, cdown;
					float nx, ny, nz, x1, y1, z1, x2, y2, z2;
					int ix, iy, iz;

					if(calc_noise) {
						NCACHE_LEFT(x, y, z) = cleft = perlin_noise_3d(fx - h, fy, fz, 2, 2, 4) * 0.5 + 0.5;
						NCACHE_RIGHT(x, y, z) = cright = perlin_noise_3d(fx + h, fy, fz, 2, 2, 4) * 0.5 + 0.5;
						NCACHE_UP(x, y, z) = cup = perlin_noise_3d(fx, fy - h, fz, 2, 2, 4) * 0.5 + 0.5;
						NCACHE_DOWN(x, y, z) = cdown = perlin_noise_3d(fx, fy + h, fz, 2, 2, 4) * 0.5 + 0.5;
					} else {
						cleft = NCACHE_LEFT(x, y, z);
						cright = NCACHE_RIGHT(x, y, z);
						cup = NCACHE_UP(x, y, z);
						cdown = NCACHE_DOWN(x, y, z);
					}
					cleft = cloud_curve(cleft, cloud_cover, cloud_sharpness);
					cright = cloud_curve(cright, cloud_cover, cloud_sharpness);
					cup = cloud_curve(cup, cloud_cover, cloud_sharpness);
					cdown = cloud_curve(cdown, cloud_cover, cloud_sharpness);

					x1 = 2.0 * h;
					y1 = 0.0;
					z1 = cright - cleft;

					x2 = 0.0;
					y2 = -2.0 * h;
					z2 = cup - cdown;

					cross(&nx, &ny, &nz, x2, y2, z2, x1, y1, z1);
					normalize(&nx, &ny, &nz);

					ix = 255.0 * (nx * 0.5 + 0.5);
					iy = 255.0 * (ny * 0.5 + 0.5);
					iz = 255.0 * (nz * 0.5 + 0.5);

					*norm++ = 0xff000000 | ((ix & 0xff) << 16) | ((iy & 0xff) << 8) | (iz & 0xff);
				}
		
				val = (int)(255.0 * cloud);
				*img++ = (val << 24) | 0xffffff;
			}
		}
	}
	fprintf(stderr, "\rdone.                           \n");

	if(calc_noise) {
		FILE *fp;
		size_t written;
		struct dump_hdr hdr;

		hdr.x = xsz;
		hdr.y = ysz;
		hdr.z = zsz;
		hdr.sdist = slice_dist_fact;

		if((fp = fopen(DUMP_FILE, "wb"))) {
			fwrite(&hdr, sizeof hdr, 1, fp);
			written = fwrite(pncache, sizeof *pncache, elem_count, fp);
			fclose(fp);

			if(written < elem_count) {
				remove(DUMP_FILE);
				perror("noise dump failed");
			} else {
				fprintf(stderr, "noise dumped sucessfully");
			}
		}

		calc_noise = 0;
	}
}
