#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>
#include <sys/time.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/glut.h>
#include "ui.h"
#include "geom.h"
#include "cgen.h"
#include "sdr.h"

#define PI					3.1415926535897931
#define TWO_PI				6.2831853071795862
#define HALF_PI				1.5707963267948966

void redraw(void);
void draw_skydome(void);
void draw_backdrop(void);
void draw_clouds(void);
void draw_sun(void);
void set_material(float r, float g, float b, float spow);
unsigned int create_texture2d(int x, int y, void *img);
unsigned int create_texture3d(int x, int y, int z, void *img);
static void recalc_cloudmap(void);
static void load_gradient(const char *fname, unsigned int *dest);
void key_handler(unsigned char key, int x, int y);
void skey_handler(int key, int x, int y);
void button_handler(int bn, int state, int x, int y);
void mouse_handler(int x, int y);
unsigned long get_msec(void);
void init_sdr(void);

int cloud_img_size = 512;
#define GRAD_IMG_SZ		128
int cloud_frames = 32;
#define SUN_IMG_SZ		128

float slice_dist_fact = 1.7;

int xsz, ysz;
int fill_polygons = 1;
float sun_phi, sun_theta;
float sun_size = 100.0;
float sun_r = 1.0, sun_g = 1.0, sun_b = 1.0;
float cam_rot, cam_pitch;
float cloud_alpha = 1.0;
float cloud_cover = 0.6;
float cloud_sharpness = 0.7;
float cloud_scale = 0.8;
int animate = 0;
int animate_clouds = 0;
float cloud_move_speed = 0.1;
float cloud_anim_speed = 0.3;
int show_image = 0;
int blend_sun = 1;

int cloudmap_invalid = 0;

unsigned int cloud_tex, cloud_norm, grad_tex, sun_tex, compass_tex;
unsigned int cloud_prog;

unsigned int msec;

int main(int argc, char **argv)
{
	int i;
	float lpos[] = {0, 1, 0, 0};

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 't':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-t must be followed by a number (texture size)\n");
					return EXIT_FAILURE;
				}
				cloud_img_size = atoi(argv[i]);
				break;

			default:
				fprintf(stderr, "unrecognized argument: %s\n", argv[i]);
				return EXIT_FAILURE;
			}
		}
	}

	if(init_gui(&argc, argv) == -1) {
		return EXIT_FAILURE;
	}

	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("clouds thingy");

	xsz = glutGet(GLUT_WINDOW_WIDTH);
	ysz = glutGet(GLUT_WINDOW_HEIGHT);

	glutDisplayFunc(redraw);
	glutKeyboardFunc(key_handler);
	glutSpecialFunc(skey_handler);
	glutMotionFunc(mouse_handler);
	glutMouseFunc(button_handler);
	glutIdleFunc(glutPostRedisplay);

	run_gui();

	init_sdr();

	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0, (float)xsz / (float)ysz, 1.0, 1000.0);

	if(!(cloud_prog = create_program_load("clouds.v.glsl", "clouds.p.glsl"))) {
		fprintf(stderr, "gamiese shader\n");
		return EXIT_FAILURE;
	}

	{
		unsigned int *img = malloc(GRAD_IMG_SZ * 4);

		load_gradient("sky.grad", img);
		
		glGenTextures(1, &grad_tex);
		glBindTexture(GL_TEXTURE_2D, grad_tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexImage2D(GL_TEXTURE_2D, 0, 4, 1, GRAD_IMG_SZ, 0, GL_BGRA, GL_UNSIGNED_BYTE, img);

		free(img);
	}

	{
		unsigned int *img = malloc(cloud_img_size * cloud_img_size * cloud_frames * 4);
		unsigned int *norm = malloc(cloud_img_size * cloud_img_size * cloud_frames * 4);

		cloud_image(img, norm, cloud_img_size, cloud_img_size, cloud_frames);

		cloud_tex = create_texture3d(cloud_img_size, cloud_img_size, cloud_frames, img);
		cloud_norm = create_texture3d(cloud_img_size, cloud_img_size, cloud_frames, norm);

		free(img);
		free(norm);
	}

	{
		int i, j;
		unsigned int *img = malloc(SUN_IMG_SZ * SUN_IMG_SZ * 4);

		for(j=0; j<SUN_IMG_SZ; j++) {
			for(i=0; i<SUN_IMG_SZ; i++) {
				float x = /*10.0 */ (float)(i - SUN_IMG_SZ / 2) / (float)(SUN_IMG_SZ / 2);
				float y = /*10.0 */ (float)(j - SUN_IMG_SZ / 2) / (float)(SUN_IMG_SZ / 2);
				float len = sqrt(x * x + y * y);
				float foo = tan(len * HALF_PI - PI / 4.0) * 0.5 + 0.5;
				float val = 1.0 - (tanh(foo * TWO_PI - PI) * 0.5 + 0.5);
				int ival = val * 255.0;

				ival = ival > 255 ? 255 : (ival < 0 ? 0 : ival);
				img[j * SUN_IMG_SZ + i] = 0xff000000 | (ival << 16) | (ival << 8) | ival;
			}
		}

		sun_tex = create_texture2d(SUN_IMG_SZ, SUN_IMG_SZ, img);

		free(img);
	}

	{
		FILE *fp;
		unsigned int *img = malloc(128 * 128 * 4);

		if(!(fp = fopen("compass.raw", "rb"))) {
			perror("failed to load compass");
			return EXIT_FAILURE;
		}
		fread(img, 4, 128 * 128, fp);
		fclose(fp);

		compass_tex = create_texture2d(128, 128, img);
		free(img);
	}

	glutMainLoop();
	return 0;
}

static float pingpong(float x)
{
	return (((int)x % 2 == 0) ? (x - floor(x)) : 1.0f - (x - floor(x)));
}

void redraw(void)
{
	msec = get_msec();

	if(cloudmap_invalid) {
		recalc_cloudmap();
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, -2, 0);
	glRotatef(cam_pitch, 1, 0, 0);
	glRotatef(cam_rot, 0, 1, 0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDisable(GL_LIGHTING);
	draw_plane(1000, 1000, 0.2, 0.6, 0.3);
	glDisable(GL_DEPTH_TEST);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, compass_tex);
	draw_plane(15, 15, 1, 1, 1);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	draw_skydome();

	if(show_image) {
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, grad_tex);

		glBegin(GL_QUADS);
		/*glColor3f(0.4, 0.4, 0.5);*/
		glTexCoord2f(0, 0); glVertex3f(-1, -1, 0);
		glTexCoord2f(1, 0); glVertex3f(1, -1, 0);
		/*glColor3f(0.1, 0.3, 1.0);*/
		glTexCoord2f(1, 1); glVertex3f(1, 1, 0);
		glTexCoord2f(0, 1); glVertex3f(-1, 1, 0);
		glEnd();

		glDisable(GL_TEXTURE_2D);

		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(0, 0, pingpong(cloud_anim_speed * (float)msec / 1000.0f) * 0.5 + 0.5);

		glEnable(GL_TEXTURE_3D);
		glBindTexture(GL_TEXTURE_3D, cloud_tex);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glBegin(GL_QUADS);
		glColor4f(1, 1, 1, cloud_alpha);
		glTexCoord2d(0, 0); glVertex3f(-1, -1, 0);
		glTexCoord2d(1, 0); glVertex3f(1, -1, 0);
		glTexCoord2d(1, 1); glVertex3f(1, 1, 0);
		glTexCoord2d(0, 1); glVertex3f(-1, 1, 0);
		glEnd();

		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_3D);

		glPopMatrix();

		glEnable(GL_LIGHTING);
		glEnable(GL_DEPTH_TEST);

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}


	glutSwapBuffers();
}


#define SKYDOME_RAD			1000.0
#define VMAX				0.2
#define SKYDOME_HEIGHT		(SKYDOME_RAD * cos(VMAX * PI))
void draw_skydome(void)
{
	glDisable(GL_LIGHTING);
	glDepthMask(0);

	draw_backdrop();
	draw_sun();
	draw_clouds();

	glDepthMask(1);
	glEnable(GL_LIGHTING);
}



void draw_backdrop(void)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(0, -SKYDOME_HEIGHT, 0);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, grad_tex);
	if(!glIsList(1)) {
		glNewList(1, GL_COMPILE_AND_EXECUTE);
		sphere(SKYDOME_RAD, 10, 1.0, VMAX, MAP_Y);
		glEndList();
	} else {
		glCallList(1);
	}
	glDisable(GL_TEXTURE_2D);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void draw_clouds(void)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(0, -SKYDOME_HEIGHT, 0);

	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glLoadIdentity();
	glScalef(cloud_scale, cloud_scale, 1.0);
	if(animate) {
		glTranslatef(0, cloud_move_speed * (float)msec / 1000.0f, 0);
	}
	if(animate_clouds) {
		glTranslatef(0, 0, pingpong(cloud_anim_speed * (float)msec / 1000.0f) * 0.5 + 0.5);
	}

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_3D);
	glBindTexture(GL_TEXTURE_3D, cloud_tex);
	glActiveTexture(GL_TEXTURE1);
	glEnable(GL_TEXTURE_3D);
	glBindTexture(GL_TEXTURE_3D, cloud_norm);

	set_uniform_int(cloud_prog, "ctex", 0);
	set_uniform_int(cloud_prog, "ntex", 1);

	bind_program(cloud_prog);

	if(!glIsList(2)) {
		glNewList(2, GL_COMPILE_AND_EXECUTE);
		sphere(SKYDOME_RAD, 10, 1.0, VMAX, MAP_XZ);
		glEndList();
	} else {
		glCallList(2);
	}

	bind_program(0);

	glActiveTexture(GL_TEXTURE1);
	glDisable(GL_TEXTURE_3D);
	glActiveTexture(GL_TEXTURE0);
	glDisable(GL_TEXTURE_3D);

	glDisable(GL_BLEND);

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void draw_sun(void)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glRotatef(sun_theta, 0, 1, 0);
	glRotatef(sun_phi, 1, 0, 0);

	if(blend_sun) glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, sun_tex);

	glBegin(GL_QUADS);
	glColor3f(sun_r, sun_g, sun_b);
	glTexCoord2f(0, 0); glVertex3f(-sun_size * 0.5, SKYDOME_RAD * 0.8, -sun_size * 0.5);
	glTexCoord2f(0, 1); glVertex3f(-sun_size * 0.5, SKYDOME_RAD * 0.8, sun_size * 0.5);
	glTexCoord2f(1, 1); glVertex3f(sun_size * 0.5, SKYDOME_RAD * 0.8, sun_size * 0.5);
	glTexCoord2f(1, 0); glVertex3f(sun_size * 0.5, SKYDOME_RAD * 0.8, -sun_size * 0.5);
	glEnd();

	glDisable(GL_TEXTURE_2D);

	glDisable(GL_BLEND);

	glPopMatrix();
}



void set_material(float r, float g, float b, float spow)
{
	float col[4];
	float spec[] = {1, 1, 1, 1};

	col[0] = r; col[1] = g; col[2] = b; col[3] = 1;

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, col);
	glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
	glMaterialf(GL_FRONT, GL_SHININESS, spow);
}

unsigned int create_texture2d(int x, int y, void *img)
{
	unsigned int tex;

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, x, y, 0, GL_BGRA, GL_UNSIGNED_BYTE, img);

	return tex;
}

unsigned int create_texture3d(int x, int y, int z, void *img)
{
	unsigned int tex;

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_3D, tex);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	glTexImage3D(GL_TEXTURE_3D, 0, 4, x, y, z, 0, GL_BGRA, GL_UNSIGNED_BYTE, img);

	return tex;
}

static void recalc_cloudmap(void)
{
	static void *img, *norm;

	if(!img) {
		img = malloc(cloud_img_size * cloud_img_size * cloud_frames * 4);
		norm = malloc(cloud_img_size * cloud_img_size * cloud_frames * 4);
	}
	cloud_image(img, norm, cloud_img_size, cloud_img_size, cloud_frames);

	glBindTexture(GL_TEXTURE_3D, cloud_tex);
	glTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, 0, cloud_img_size, cloud_img_size, cloud_frames, GL_BGRA, GL_UNSIGNED_BYTE, img);

	cloudmap_invalid = 0;
}

struct grad_key {
	int pos;
	float r, g, b;
};

#define LERP(a, b, t)	((a) + ((b) - (a)) * (t))
static void load_gradient(const char *fname, unsigned int *dest)
{
	FILE *fp;
	char buf[256], *ptr;
	struct grad_key keys[100];
	int i, knum = 0;

	if(!(fp = fopen(fname, "r"))) {
		perror("failed to open grad file");
		return;
	}

	while(fgets(buf, sizeof buf, fp)) {
		assert(ptr = strtok(buf, " \n\r\t"));
		keys[knum].pos = (float)GRAD_IMG_SZ * (float)atoi(ptr) / 100.0;
		assert(ptr = strtok(0, " \n\r\t"));
		keys[knum].r = atof(ptr);
		assert(ptr = strtok(0, " \n\r\t"));
		keys[knum].g = atof(ptr);
		assert(ptr = strtok(0, " \n\r\t"));
		keys[knum].b = atof(ptr);
		knum++;
	}
	fclose(fp);

	for(i=0; i<keys[0].pos; i++) {
		int r = keys[0].r * 255.0;
		int g = keys[0].g * 255.0;
		int b = keys[0].b * 255.0;
		*dest++ = r << 16 | g << 8 | b;
	}

	for(i=0; i<knum-1; i++) {
		int j, span = keys[i+1].pos - keys[i].pos;
		for(j=0; j<span; j++) {
			float t = (float)j / (float)span;
			int r = LERP(keys[i].r, keys[i+1].r, t) * 255.0;
			int g = LERP(keys[i].g, keys[i+1].g, t) * 255.0;
			int b = LERP(keys[i].b, keys[i+1].b, t) * 255.0;
			*dest++ = r << 16 | g << 8 | b;
		}
	}

	for(i = keys[knum-1].pos; i<GRAD_IMG_SZ; i++) {
		int r = keys[knum-1].r * 255.0;
		int g = keys[knum-1].g * 255.0;
		int b = keys[knum-1].b * 255.0;
		*dest++ = r << 16 | g << 8 | b;
	}
}

void key_handler(unsigned char key, int x, int y)
{
	skey_handler(key, x, y);
}

void skey_handler(int key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case ' ':
		fill_polygons = !fill_polygons;
		glPolygonMode(GL_FRONT_AND_BACK, fill_polygons ? GL_FILL : GL_LINE);
		break;

	case GLUT_KEY_PAGE_UP:
		cloud_alpha += 0.05;
		break;

	case GLUT_KEY_PAGE_DOWN:
		cloud_alpha -= 0.05;
		break;

	case '`':
		show_image = !show_image;
		break;

	case 's':
		blend_sun = !blend_sun;
		break;
	}
}

int prev_x = -1;
int prev_y = -1;

void button_handler(int button, int state, int x, int y)
{
	if(button == GLUT_RIGHT_BUTTON) {
		if(state == GLUT_DOWN) {
			prev_x = x;
			prev_y = y;
		} else {
			prev_x = -1;
			prev_y = -1;
		}
	}
}

void mouse_handler(int x, int y)
{
	if(prev_x != -1) {
		cam_rot += (float)(x - prev_x) / 2.0f;
		prev_x = x;
	}

	if(prev_y != -1) {
		cam_pitch += (float)(y - prev_y) / 2.0f;
		/*if(cam_pitch > 0.0) cam_pitch = 0.0;*/
		/*if(cam_pitch < -90.0) cam_pitch = -90.0;*/
		prev_y = y;
	}
}

unsigned long get_msec(void)
{
	static struct timeval timeval, first_timeval;

	gettimeofday(&timeval, 0);

	if(first_timeval.tv_sec == 0) {
		first_timeval = timeval;
		return 0;
	}
	return (timeval.tv_sec - first_timeval.tv_sec) * 1000 + (timeval.tv_usec - first_timeval.tv_usec) / 1000;
}
