#include <math.h>
#include <GL/gl.h>
#include "sdr.h"
#include "geom.h"

#define PI			3.1415926535897931
#define HALF_PI		1.5707963267948966
#define TWO_PI		6.2831853071795862

void draw_plane(float w, float h, float r, float g, float b)
{
	w /= 2;
	h /= 2;

	glBegin(GL_QUADS);
	glColor3f(r, g, b);
	glNormal3f(0, 1, 0);
	glTexCoord2f(0, 0); glVertex3f(-w, 0, -h);
	glTexCoord2f(1, 0); glVertex3f(w, 0, -h);
	glTexCoord2f(1, 1); glVertex3f(w, 0, h);
	glTexCoord2f(0, 1); glVertex3f(-w, 0, h);
	glEnd();
}

#define PT0		{0, 1, 0}
#define PT1		{0, -1, 0}
#define PT2		{0.707, 0, 0.707}
#define PT3		{-0.707, 0, 0.707}
#define PT4		{-0.707, 0, -0.707}
#define PT5		{0.707, 0, -0.707}

static vec3_t sphv[] = {
	PT0, PT5, PT2,
	PT0, PT2, PT3,
	PT0, PT3, PT4,
	PT0, PT4, PT5,
	PT1, PT2, PT5,
	PT1, PT3, PT2,
	PT1, PT4, PT3,
	PT1, PT5, PT4,
};

static void gsphere_triangle(vec3_t a, vec3_t b, vec3_t c, int iter, int mapping);

void gsphere(float sz, int iter, int hemisphere, int mapping)
{
	int i, vcount = hemisphere ? 4 : 8;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glScalef(sz, sz, sz);

	glEnable(GL_NORMALIZE);

	glBegin(GL_TRIANGLES);

	for(i=0; i<vcount; i++) {
		gsphere_triangle(sphv[i * 3], sphv[i * 3 + 1], sphv[i * 3 + 2], iter, mapping);
	}

	glEnd();

	glDisable(GL_NORMALIZE);

	glPopMatrix();
}

static void gsphere_vertex(float x, float y, float z, int mapping)
{
	if(mapping == MAP_XZ) {
		float dist = sqrt(x * x + y * y + z * z);
		float newdist = sin(dist * HALF_PI - HALF_PI) + 1.0;

		float u = newdist * x / dist;
		float v = newdist * z / dist;

		glTexCoord2f(u, v);
	} else {
		glTexCoord2f(0.0, y);
	}
	glNormal3f(x, y, z);
	glVertex3f(x, y, z);
}

static void gsphere_triangle(vec3_t a, vec3_t b, vec3_t c, int iter, int mapping)
{
	if(iter-- > 0) {
		vec3_t d, e, f;
		float dlen, elen, flen;

		d.x = a.x + c.x;
		d.y = a.y + c.y;
		d.z = a.z + c.z;
		dlen = sqrt(d.x * d.x + d.y * d.y + d.z * d.z);
		d.x /= dlen;
		d.y /= dlen;
		d.z /= dlen;

		e.x = a.x + b.x;
		e.y = a.y + b.y;
		e.z = a.z + b.z;
		elen = sqrt(e.x * e.x + e.y * e.y + e.z * e.z);
		e.x /= elen;
		e.y /= elen;
		e.z /= elen;

		f.x = b.x + c.x;
		f.y = b.y + c.y;
		f.z = b.z + c.z;
		flen = sqrt(f.x * f.x + f.y * f.y + f.z * f.z);
		f.x /= flen;
		f.y /= flen;
		f.z /= flen;

		gsphere_triangle(a, e, d, iter, mapping);
		gsphere_triangle(d, f, c, iter, mapping);
		gsphere_triangle(e, b, f, iter, mapping);
		gsphere_triangle(e, f, d, iter, mapping);
	} else {
		gsphere_vertex(a.x, a.y, a.z, mapping);
		gsphere_vertex(b.x, b.y, b.z, mapping);
		gsphere_vertex(c.x, c.y, c.z, mapping);
	}
}


static void cross(float *rx, float *ry, float *rz, float x1, float y1, float z1, float x2, float y2, float z2)
{
	*rx = y1 * z2 - z1 * y2;
	*ry = z1 * x2 - x1 * z2;
	*rz = x1 * y2 - y1 * x2;
}

static int attr_tangent;

static void sphere_vertex(float x, float y, float z, float rad, float u, float v, float vmax, int mapping)
{
	float len, nx, ny, nz, tx, ty, tz, ux = 0, uy = 1, uz = 0;
	
	/* tex coords */
	if(mapping == MAP_XZ) {
		glTexCoord2f(x / rad, z / rad);
	} else if(mapping == MAP_Y) {
		float lowest = rad * cos(vmax * PI);
		float height = rad - lowest;

		glTexCoord2f(0, (y - lowest) / height);
	} else {
		glTexCoord2f(u, v);
	}

	/* normal */
	nx = x / rad;
	ny = y / rad;
	nz = z / rad;
	glNormal3f(-nx, -ny, -nz);	/* inside-looking */

	/* calc tangent */
	cross(&tx, &ty, &tz, nx, ny, nz, ux, uy, uz);
	len = sqrt(tx * tx + ty * ty + tz * tz);
	set_attrib_float3(attr_tangent, tx / len, ty / len, tz / len);
	
	glVertex3f(x, y, z);
}

#define SPH_X(rho, theta, phi)	((rho) * cos(theta) * sin(phi))
#define SPH_Y(rho, theta, phi)	((rho) * cos(phi))
#define SPH_Z(rho, theta, phi)	((rho) * sin(theta) * sin(phi))

void sphere(float rad, int vpoints, float umax, float vmax, int mapping)
{
	unsigned int cur_prog;
	float u0, u1, v0, v1;
	float du, dv;
	int i, j, upoints = vpoints * 2;

	du = umax / upoints;
	dv = vmax / vpoints;

	v0 = 0.0;
	v1 = dv;

	glGetIntegerv(GL_CURRENT_PROGRAM, &cur_prog);
	attr_tangent = get_attrib_loc(cur_prog, "vtang");

	glBegin(GL_QUADS);
	glColor3f(1, 1, 1);
	for(i=0; i<vpoints; i++) {
		float phi0 = v0 * PI;
		float phi1 = v1 * PI;
	
		u0 = 0.0;
		u1 = du;

		for(j=0; j<upoints; j++) {
			float theta0 = u0 * TWO_PI;
			float theta1 = u1 * TWO_PI;

			sphere_vertex(SPH_X(rad, theta0, phi0), SPH_Y(rad, theta0, phi0), SPH_Z(rad, theta0, phi0), rad, u0, v0, vmax, mapping);
			sphere_vertex(SPH_X(rad, theta1, phi0), SPH_Y(rad, theta1, phi0), SPH_Z(rad, theta1, phi0), rad, u1, v0, vmax, mapping);
			sphere_vertex(SPH_X(rad, theta1, phi1), SPH_Y(rad, theta1, phi1), SPH_Z(rad, theta1, phi1), rad, u1, v1, vmax, mapping);
			sphere_vertex(SPH_X(rad, theta0, phi1), SPH_Y(rad, theta0, phi1), SPH_Z(rad, theta0, phi1), rad, u0, v1, vmax, mapping);

			u0 += du;
			u1 += du;
		}
		v0 += dv;
		v1 += dv;
	}
	glEnd();
}

void overlay(float l, float t, float r, float b, float alpha, unsigned int tex)
{
	l = l * 2.0 - 1.0;
	t = t * 2.0 - 1.0;
	r = r * 2.0 - 1.0;
	b = b * 2.0 - 1.0;

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glBindTexture(GL_TEXTURE_2D, tex);
	glEnable(GL_TEXTURE_2D);

	glBegin(GL_QUADS);
	glColor4f(1, 1, 1, alpha);
	glTexCoord2d(0, 0); glVertex3f(l, t, 0);
	glTexCoord2d(1, 0); glVertex3f(r, t, 0);
	glTexCoord2d(1, 1); glVertex3f(r, b, 0);
	glTexCoord2d(0, 1); glVertex3f(l, b, 0);
	glEnd();

	glDisable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

